#!/bin/perl

### copyright 2006 by Alan Moses. This software is provided "as is" without
### warranty of any kind.  The author assumes no responsibility for the results it
### produces or conclusions based thereupon.  It is distributed free of charge for
### academic use only.  Permission to copy and use it is granted free of charge
### provided that no fee is charged and this copyright notice is not removed.

my $MESSAGES=1;
my %LOGFACTCOMPUTED;
die "usage : [fasta file] [motif(=0.XXXX)] or try -help for more options\n" unless (@ARGV);

if ($ARGV[0] =~m/\-h/) { help(); exit; }

my $fastafile = shift @ARGV;

my %SEQ; die "$fastafile not read \n" unless (read_fasta(\%SEQ,$fastafile));

my %userBGF;
for my $i (0 .. scalar(@ARGV)-1) { 
	my $m = $ARGV[$i];
	if ($m =~ m/(\S+)=(0\.\d+)/) {
		my $c=consensus2regex($1);
		$userBGF{$c}=$2;
		$ARGV[$i]=$c;
	}	
}

if ($MESSAGES) {
	print STDERR scalar(@ARGV)," motifs found "; 
	if (%userBGF) { for my $m (@ARGV) { print STDERR "$m: $userBGF{$m} "; } }
	print STDERR "\n", scalar keys %SEQ, " sequences found\n";
}

my %occ; my %freq; my %type; my %total; my $totalaa=0; my %size;
for my $s (@{$SEQ{"names"}}) {
	$SEQ{$s} =~ s/\*//g;
	$freq{$s}{$motif}="0";
	my %covered;
	for my $motif (@ARGV) {
	while ($SEQ{$s} =~ m/($motif)/g) {
		my $pos = pos($SEQ{$s})-length($1)+1; #one based coordinates
		my $match=$1;
		if ($type{$s}{$pos} eq "") {
			$type{$s}{$pos}=$motif;
			push @{$occ{$s}}, $pos;  
			$total{$motif}++;
			$freq{$s}{$motif}++;
			for my $p (0 .. length($match)-1) { $covered{$pos+$p}=1; }
		}
	}
	}
	$totalaa += length ($SEQ{$s}); 
	next unless (@{$occ{$s}}); #only count sequences that have at least one match 
	
#set the 'size' (waiting time) for each match, circularizing proteins
	if (scalar(@{$occ{$s}})==1) { $size{$s}{$occ{$s}[0]}=length($SEQ{$s}); }
	else {
		@{$occ{$s}} = sort numerically @{$occ{$s}};
		$size{$s}{$occ{$s}[0]}=$occ{$s}[0]+length($SEQ{$s})-$occ{$s}[-1]-1;
		for my $h (0 .. scalar(@{$occ{$s}})-2) {			
			$size{$s}{$occ{$s}[$h+1]} = $occ{$s}[$h+1] - $occ{$s}[$h];
		}		
	}	
}

my %bgf;

for my $motif (@ARGV) {
	$bgf{$motif} = ($userBGF{$motif} eq "") ? $total{$motif}/$totalaa : $userBGF{$motif};
	print STDERR "$motif: $total{$motif} matches in $totalaa aa (f=$bgf{$motif})\n" if ($MESSAGES);
}
print "SEQ\tNAME\t";
print "SCORE\tPVAL\tCORR\t"; for my $m (@ARGV) { print $m."\t"; } 
print "Lcl\t"; for my $m (@ARGV) { print $m."\t"; } print "Lseq\tpos\tfoldedness\n";
for my $s (@{$SEQ{"names"}}) {
	my $bests=0; my $bestl; my $bestr; my %bestcl;
	for my $i ( 0 .. scalar(@{$occ{$s}})-1) {
		CL: for my $j ($i+1 .. scalar(@{$occ{$s}})-1) {
			my %cl; my $prod=1;
			for my $k ($i .. $j) { $cl{$type{$s}{$occ{$s}[$k]}}++; }
			my %uvp;
			for my $m (@ARGV) {			
				my $n=$cl{$m};	
				$uvp{$m}=($n>0) ? 1-binomial($n-1, $occ{$s}[$j]-$occ{$s}[$i], $bgf{$m}): 1;
			}
			my $p=0;
			for my $m (@ARGV) { $p -= log($uvp{$m}); }
			if ($p>$bests) { #store the best scoring cluster
				$bestl=$i; $bestr=$j; $bests=$p; %bestcl=%cl; 	
			}
			elsif (($p==$bests) && (($occ{$s}[$bestr]-$occ{$s}[$bestl])>($occ{$s}[$j]-$occ{$s}[$i]))) {
				$bestl=$i; $bestr=$j; $bests=$p; %bestcl=%cl; 
				
				
			}
		}
	}
	my $corrscore=0.0; my $clscore=0.0;
	my $corr = 0.5*scalar(@{$occ{$s}})*(scalar(@{$occ{$s}})-1);
	if ($bests!=0) { 
		my $rawp=pval_prod_pval(exp(-$bests),scalar(@ARGV)); #Q fast algorithm
		$corrscore = -log($corr*$rawp); #correct for multiple testing
		$clscore = -log($rawp);
	}

	my $gname;
	if ($s =~ m/(ENS\w*G\d+)/) { $gname=$1; }
	else { $gname = $1 if ($s =~m/^(\S+)\s/); }
	print "$s\t$gname\t$bests\t$clscore\t$corrscore\t";
	for my $m (@ARGV) { print $bestcl{$m}."\t"; }	
	print ($occ{$s}[$bestr]-$occ{$s}[$bestl]); print "\t";
	for my $m (@ARGV) { print $freq{$s}{$m}."\t"; }
	my $midp;
	if ($bests!=0) { $midp=$occ{$s}[$bestl]."-".$occ{$s}[$bestr]; }
	else { $midp="NF"; }
	print length($SEQ{$s})."\t$midp\t";
	if ($bests!=0) { print fold_index(substr($SEQ{$s},$occ{$s}[$bestl],$occ{$s}[$bestr]-$occ{$s}[$bestl])); }
	else { print "NF"; }
	print "\n";
}

sub help {	

	print "perl script to compute the S_bn statistic\nv1.0 by Alan Moses 2006\n";
	print "\nusage : [fasta file] [motif(=0.XXXX)]\n";
	print "\noutput\n",
		"\nPrints tab delimited to standard output.\n";
 	print "\nfasta file\n",
		"\nThe first argument is a multiple fasta file.  The header lines\n",
		"start with a '>' symbol.  There there is no minimum or maximum\n","line length.\n";
	print "\nmotifs\n",
		"\nThe remaining arguments are motifs. These can be simple perl\n",
		"regular expressions or simple strings.  \[ST\]P (i.e., S or T \n",
		"followed by P) should be entered as \'[ST]P\'. The program also\n",
		"recognizes X or B as any leter or MILVAPW, respectively. These\n",
		"can be entered as RXXL or \'RXXL\'. In some cases the user may\n",
		"want to specify the background frequencies of the motifs.  This\n",
		"can be done by entering and = followed by the frequency, per\n",
		"amino acid residue. E.g.,\n\n",
		"% perl compute_SLR.pl file.fasta \'[ST]P\'=0.001 \'RXXL\'=0.001 > test\n\n",
		"looks for clusters of S/T-P and R-X-X-L relative to background\n",
		"expectation of 1 instance per 1000 amino acids for both motifs\n",
		"and writes the output to a file called test.\n";

	
	

}
sub fold_index { #as in Prilusky et al 2005
	my ($seq) = @_;
	$seq =~ s/X//g;
	return (2.785*avg_hydropathy($seq) - abs(avg_charge($seq)) - 1.151);
}

sub avg_charge {#as in Uversky et al. 2000
	my ($seq) = @_;	
	#my %ch = ( "R"=>"1", "K"=>"1", "D"=>"-1", "E"=>"-1" ); #as in Uversky et al. 2000
	my $c=0;
	$c++ while ($seq =~ m/[RK]/g);
	$c-- while ($seq =~ m/[DE]/g);
	return ($c/length($seq));
}

sub avg_hydropathy { #as in Uversky et al. 2000 - scale between 0 and 1 by dividing by 9.0
	my ($seq) = @_;
	my $h=0;
	my %HP = ( #Kyte & Doolittle 1982
"R"=>"0.0","K"=>"0.6","D"=>"1.0","B"=>"1.0","N"=>"1.0","S"=>"3.6","E"=>"1.0",
"H"=>"1.3","Z"=>"1.0","Q"=>"1.0","T"=>"3.8","G"=>"4.1","X"=>"4.1","A"=>"6.3",
"P"=>"2.9","V"=>"8.7","Y"=>"3.2","C"=>"7.0","M"=>"6.4","I"=>"9.0","L"=>"8.2",
"W"=>"3.6","F"=>"7.2");

	$h+=($HP{$1}/9.0) while ($seq =~ m/([A-Z])/g);
	return ($h/length($seq));
}
sub pval_prod_pval { #Qfast algorithm
	my ($p,$n)=@_;
	#print "-->($p,$n\n";
	return 0 if ($p==0);
	my $logp=-log($p);
	my $pval=$p; my $thisp=$p;
	for $i (1 .. $n-1) {
		$thisp*=($logp/$i);
		$pval+=$thisp;
	}	
	#print "$p $n $pval\n";
	return $pval;
}

sub binomial { #probability of k or fewer of $n when you expect $p x n
        my ($k,$n,$p) = @_;
        my $pval=0; #if ($p==0) { die "problem with $k $n $p\n"; }
        for (my $i=0;$i<=$k;$i++) {
                $pval += exp(logchoos($n,$i) + $i*log($p) + ($n-$i)*log(1.0-$p));
        }
        if ($pval>=1.0) { print STDERR "warning: $k $n $p $pval\n"; $pval=0.999999999999; }
        return $pval;
 
}
sub logchoos {
        #compute log "n choose k" or ( n ) = n!/(n-k)!k!
        #                            ( k )
        my ($n,$k) = @_;
 
        return (logfact($n) - logfact($n-$k) - logfact($k));
 
}
 
sub logfact {
        my ($n) = @_;
        return 0 if ($n<=1);
        return ($LOGFACTCOMPUTED{$n}) if ($LOGFACTCOMPUTED{$n});
        return ($LOGFACTCOMPUTED{$n}=log($n) + logfact($n-1));
}



sub consensus2regex {
	my ($m) = (@_);
	$m=~ s/X/[A-Z]/g;
	$m=~ s/B/[MILVAPW]/g;
	return $m;
}

sub read_fasta {
	my ($data, $file) = @_;
	open (FILE, "<", $file) or return 0;
	my $name;
	while (my $line=<FILE>) { chomp $line;
		if ($line =~ m/>(.+)/) {
			$name=$1; 
			push @{$$data{"names"}}, $name;			
		}
		else { $$data{$name}.=$line; }
	}
	return 1;
}

sub numerically { $a <=> $b }

