#!/bin/perl

### copyright 2006 by Alan Moses. This software is provided "as is" without
### warranty of any kind.  The author assumes no responsibility for the results it
### produces or conclusions based thereupon.  It is distributed free of charge for
### academic use only.  Permission to copy and use it is granted free of charge
### provided that no fee is charged and this copyright notice is not removed.

my $MESSAGES=1; my $MIDP=0; my $R2L=0;

die "usage : [fasta file] [motif(=0.XXXX)] or try -help for more options\n" unless (@ARGV);

if ($ARGV[0] =~m/\-h/) { help(); exit; }

my $fastafile = shift @ARGV;

my %SEQ; die "$fastafile not read \n" unless (read_fasta(\%SEQ,$fastafile));

my %userBGF;
for my $i (0 .. scalar(@ARGV)-1) { 
	my $m = $ARGV[$i];
	if ($m =~ m/(\S+)=(0\.\d+)/) {
		my $c=consensus2regex($1);
		$userBGF{$c}=$2;
		$ARGV[$i]=$c;
	}	
}

if ($MESSAGES) {
	print STDERR scalar(@ARGV)," motifs found "; 
	if (%userBGF) { for my $m (@ARGV) { print STDERR "$m: $userBGF{$m} "; } }
	print STDERR "\n", scalar keys %SEQ, " sequences found\n";
}

my %occ; my %freq; my %type; my %total; my $totalaa=0; my %size;
for my $s (@{$SEQ{"names"}}) {
	$SEQ{$s} =~ s/\*//g;
	$freq{$s}{$motif}="0";
	my %covered;
	for my $motif (@ARGV) {
	while ($SEQ{$s} =~ m/($motif)/g) {
		my $pos = pos($SEQ{$s})-length($1)+1; #one based coordinates
		my $match=$1;
		if ($covered{$pos}!=1) {
			$type{$s}{$pos}=$motif;
			push @{$occ{$s}}, $pos;  
			$total{$motif}++;
			$freq{$s}{$motif}++;
			for my $p (0 .. length($match)-1) { $covered{$pos+$p}=1; }
		}
	}
	}
	next unless (@{$occ{$s}}); #only count sequences that have at least one match 
	$totalaa += length ($SEQ{$s}); 
#set the 'size' (waiting time) for each match
	if (scalar(@{$occ{$s}})==1) { $size{$s}{$occ{$s}[0]}=length($SEQ{$s}); }
	elsif ($MIDP==1) {
		@{$occ{$s}} = sort numerically @{$occ{$s}};
		$size{$s}{$occ{$s}[0]}= $occ{$s}[0] + 0.5*($occ{$s}[1]-$occ{$s}[0]);
		$size{$s}{$occ{$s}[-1]}= length($SEQ{$s})-$occ{$s}[-1] + 0.5*($occ{$s}[-1]-$occ{$s}[-2]);
		for my $h (0 .. scalar(@{$occ{$s}})-3) {			
			$size{$s}{$occ{$s}[$h+1]} = 0.5*($occ{$s}[$h+1] - $occ{$s}[$h]) + 0.5*($occ{$s}[$h+2] - $occ{$s}[$h+1]);
		}		
	}
	elsif ($R2L==1) {
		@{$occ{$s}} = sort numerically @{$occ{$s}};		
		$size{$s}{$occ{$s}[-1]}= length($SEQ{$s})-$occ{$s}[-1] + $occ{$s}[0];
		for my $h (0 .. scalar(@{$occ{$s}})-2) {			
			$size{$s}{$occ{$s}[$h]} = $occ{$s}[$h+1] - $occ{$s}[$h];
		}		
	}
	else { 
		@{$occ{$s}} = sort numerically @{$occ{$s}};		
		$size{$s}{$occ{$s}[0]}= $occ{$s}[0]+length($SEQ{$s})-$occ{$s}[-1]-1;
		for my $h (0 .. scalar(@{$occ{$s}})-2) {			
			$size{$s}{$occ{$s}[$h+1]} = $occ{$s}[$h+1] - $occ{$s}[$h];
		}
	}	
}

my %bgf;

for my $motif (@ARGV) {
	$bgf{$motif} = ($userBGF{$motif} eq "") ? $total{$motif}/$totalaa : $userBGF{$motif};
	print STDERR "$motif: $total{$motif} matches in $totalaa aa (f=$bgf{$motif})\n" if ($MESSAGES);
}

print "SEQ\tNAME\t"; for my $m (@ARGV) { print $m."\tfc($m)\tE[N]\t"; }
print "log(LRc)\tLseq\tE[l]\tE[pos]\tlog(LRns)\tSlr\n";
for my $s (@{$SEQ{"names"}}) {
	my $gname;
	if ($s =~ m/(ENS\w*G\d+)/) { $gname=$1; }
	else { $gname = $1 if ($s =~m/^(\S+)\s/); }
	print "$s\t$gname\t";
	my $pri; my $llh=0; my $cllh=0;
	if (scalar(@{$occ{$s}})>=1) {
		$llh=0;
		my @temp;
		for my $start ( 1 .. 5) {
			my %inclf;
			for my $m (@ARGV) { 				
				$inclf{$m} = $bgf{$m};
				$inclf{$m} *= (1.0+rand) if ($m eq $ARGV[0]);
				 #initialize at background frequencies with a random deviation.
				$pri= 0.5*(0.5+rand); #initialize at 1/2 with a random deviation.
			}	
			@temp= fit_geomix_fc_pri(\@{$occ{$s}},\%{$type{$s}},\%{$size{$s}},\%inclf,\%bgf,$pri,\@ARGV); #fit the cluster densities and the prior.
			if ($temp[1]>=$llh) { #take the best of the 5 starts
				$pri=shift @temp;  $llh = shift @temp; 
				for my $m (@ARGV) {  $clf{$m}=shift @temp; }
			}
		}
		if (scalar(@ARGV)>1) { #compute the non-specific cluster 'control' likelihood - the first motif is assumed to be the more specific one
			for my $start ( 1 .. 5) {
				my %inclf;
				for my $m (@ARGV) { 
					
					$inclf{$m} = $bgf{$m};
					$inclf{$m} *= (1.0+rand) if ($m eq $ARGV[0]);
				}
				$pri= 0.5*(0.5+rand);
				@temp= fit_geomix_control_pri(\@{$occ{$s}},\%{$type{$s}},\%{$size{$s}},\%inclf,\%bgf,$pri,\@ARGV);
				
				if ($temp[1]>=$cllh) { 
					my $junk=shift @temp;  $cllh = shift @temp; 
				}
			}
		}
		#compute some fun quantities
		my %EN;  my $EL=0; my $totpos=0; my $totN=scalar(@{$occ{$s}});
		for my $pos (@{$occ{$s}}) {
			my $llr = mv_geom_llr($type{$s}{$pos},$size{$s}{$pos},\%clf,\%bgf);
			my $temp2 = $pri*exp($llr);
			my $post = ($temp2>0) ? 1/( 1 + (1-$pri)/$temp2 ) : 0;
			$EN{$type{$s}{$pos}}+=$post; 
			$EL+=($post*$size{$s}{$pos});
			$totpos+=$post*$pos;
		}
		for my $m (@ARGV) { 			
			print "$freq{$s}{$m}\t$clf{$m}\t$EN{$m}\t"; 
		}
		my $midp=($totN>0)?($totpos/$totN):"NA";
		print "$llh\t".length($SEQ{$s})."\t$EL\t$midp\t$cllh\t", log(exp($llh)/(1+exp($cllh))),"\n";
	}
	else {
		for my $m (@ARGV) { 			
			print "$freq{$s}{$m}\tNA\tNA\t"; 
		}
		print "0\t".length($SEQ{$s})."\tNA\tNA\t0\t",log(exp(0)/(1+exp(0))),"\n";
	}
}

sub help {	

	print "perl script to compute the S_lr statistic\nv1.01 by Alan Moses 2006\n";
	print "\nusage : [fasta file] [motif(=0.XXXX)]\n";
	print "\noutput\n",
		"\nPrints tab delimited to standard output.\n";
 	print "\nfasta file\n",
		"\nThe first argument is a multiple fasta file.  The header lines\n",
		"start with a '>' symbol.  There there is no minimum or maximum\n","line length.\n";
	print "\nmotifs\n",
		"\nThe remaining arguments are motifs. These can be simple perl\n",
		"regular expressions or simple strings.  \[ST\]P (i.e., S or T \n",
		"followed by P) should be entered as \'[ST]P\'. The program also\n",
		"recognizes X or B as any leter or MILVAPW, respectively. These\n",
		"can be entered as RXXL or \'RXXL\'. In some cases the user may\n",
		"want to specify the background frequencies of the motifs.  This\n",
		"can be done by entering and = followed by the frequency, per\n",
		"amino acid residue. E.g.,\n\n",
		"% perl compute_SLR.pl file.fasta \'[ST]P\'=0.001 \'RXXL\'=0.001 > test\n\n",
		"looks for clusters of S/T-P and R-X-X-L relative to background\n",
		"expectation of 1 instance per 1000 amino acids for both motifs\n",
		"and writes the output to a file called test.\n";

	
	

}


sub fit_geomix_control_pri {
	my ($occ,$type,$size,$ifc,$fb,$ipri,$mot)=@_;
	my %fc=%{$ifc}; 
	my $initllh=0; my $lastp=0; my $pri=$ipri;
	for my $p (@{$occ}) { #compute the starting loglikelihood
		$initllh+=log(geomix_lh($$type{$p},$fb,$fb,$pri,$$size{$p},$mot)); 
	}
	my $imax=100; #maximum number of em iterations
	my $ep = 0.001;  #convergence criterion
	my $lastlh; my $counter=0;
	for my $i (1 .. $imax) {
		my %pos; my $llh=0; my $EC=0; my %post; my %ENc; my $EL=0;
		$lastp=0;
		for my $p (@{$occ}) {			
			my $temp = $pri*exp(mv_geom_llr($$type{$p},$$size{$p},\%fc,$fb));
			#posterior only depends on the likelihood ratio, as long as it's greater than zero
			$post{$p} = ($temp>0) ? 1/( 1 + (1-$pri)/$temp) : 0; 	
			$ENc{$$type{$p}} += $post{$p}; 	#expected number of each type
			$EC+=$post{$p}; #expected number in 'cluster' class
			$EL += $post{$p}*$$size{$p}; #expected length in 'cluster' class
			$llh+=log(geomix_lh($$type{$p},\%fc,$fb,$pri,$$size{$p},$mot));
			
		}
		
		$pri = $EC/scalar(@{$occ}); #update prior
		my @fvals;
		for my $m (@{$mot}) { 
			$fc{$m} = (($ENc{$m}/$EL)>0.00001) ? $ENc{$m}/$EL : 0.00001;
			$fc{$m} = 0.1 if ($fc{$m}>0.1);
			#update frequencies, make sure they aren't too small or too big			
			$fc{$m} = $$fb{$m} if (($m eq $$mot[0])&&($fc{$m}>$$fb{$m})); ###this is the only difference - constrain the frequency of the first motif to be less than the background###
			push @fvals, $fc{$m};
		}
		#decide to restart, continue or return	
		if (($pri!=0)&&($counter>20)&&($llh-$initllh<$ep)) { 
			$pri = rand ; $counter=0; 
		}
		elsif ((($i>3)&&($llh-$lastlh < $ep)&&($llh-$initllh>=0))||($imax==$i)) { 
			return ($pri,$llh-$initllh,@fvals);
		}
		$lastlh=$llh;	
		$counter++;		
	}
}
sub fit_geomix_fc_pri {
	my ($occ,$type,$size,$ifc,$fb,$ipri,$mot)=@_;
	my %fc=%{$ifc}; 
	my $initllh=0; my $lastp=0; my $pri=$ipri;
	for my $p (@{$occ}) { #compute the starting loglikelihood
		$initllh+=log(geomix_lh($$type{$p},$fb,$fb,$pri,$$size{$p},$mot)); 
	}
	my $imax=100; #maximum number of em iterations
	my $ep = 0.001;  #convergence criterion
	my $lastlh; my $counter=0;
	for my $i (1 .. $imax) {
		my %pos; my $llh=0; my $EC=0; my %post; my %ENc; my $EL=0;
		$lastp=0;
		for my $p (@{$occ}) {			
			my $temp = $pri*exp(mv_geom_llr($$type{$p},$$size{$p},\%fc,$fb));
			#posterior only depends on the likelihood ratio, as long as it's greater than zero
			$post{$p} = ($temp>0) ? 1/( 1 + (1-$pri)/$temp) : 0; 	
			$ENc{$$type{$p}} += $post{$p}; 	#expected number of each type
			$EC+=$post{$p}; #expected number in 'cluster' class
			$EL += $post{$p}*$$size{$p}; #expected length in 'cluster' class
			$llh+=log(geomix_lh($$type{$p},\%fc,$fb,$pri,$$size{$p},$mot));
			
		}
		
		$pri = $EC/scalar(@{$occ}); #update prior
		my @fvals;
		for my $m (@{$mot}) { 
			$fc{$m} = (($ENc{$m}/$EL)>0.00001) ? $ENc{$m}/$EL : 0.00001;
			$fc{$m} = 0.1 if ($fc{$m}>0.1);
			#update frequencies, make sure they aren't too small or too big			
			push @fvals, $fc{$m};
		}
		#decide to restart, continue or return	
		if (($pri!=0)&&($counter>20)&&($llh-$initllh<$ep)) { 
			$pri = rand ; $counter=0; 
		}
		elsif ((($i>3)&&($llh-$lastlh < $ep)&&($llh-$initllh>=0))||($imax==$i)) { 
			return ($pri,$llh-$initllh,@fvals);
		}
		$lastlh=$llh;	
		$counter++;		
	}
}

sub geomix_lh { #likelihood of an observation under a two-component mixture of geometric
	my ($type,$fc,$fb,$pri,$len,$mot)=@_;
	my $totbf=1; my $totcf=1;
	for my $m (@{$mot}) { $totbf-= $$fb{$m};  $totcf -= $$fc{$m}; }	
	my $lh = $pri*($$fc{$type}*exp(($len-1)*log($totcf)))
			+(1.0-$pri)*($$fb{$type}*exp(($len-1)*log($totbf))); #do the powers by logs and exponentials
	return $lh;		
}

sub mv_geom_llr { #log likelihood ratio of an observation 
	my ($type,$len,$fc,$fb) = @_;	
	my $ctot=1; my $btot=1;
	for my $m (keys %{$fb}) { $btot-=$$fb{$m}; $ctot-=$$fc{$m}; }	
	return (log($$fc{$type}/$$fb{$type}) + ($len-1)*log($ctot/$btot));	
}
sub consensus2regex {
	my ($m) = (@_);
	$m=~ s/X/[A-Z]/g;
	$m=~ s/B/[MILVAPW]/g;
	return $m;
}

sub read_fasta {
	my ($data, $file) = @_;
	open (FILE, "<", $file) or return 0;
	my $name;
	while (my $line=<FILE>) { chomp $line;
		if ($line =~ m/>(.+)/) {
			$name=$1; 
			push @{$$data{"names"}}, $name;			
		}
		else { $$data{$name}.=$line; }
	}
	return 1;
}

sub numerically { $a <=> $b }

